import logging
from queue import Queue
from threading import Thread

logger = logging.getLogger('analyser.reactor')


class Reactor(object):
    """
    Kicks off the thread which accepts requests on the queue where each request is a 2 item tuple ( request type,
    tuple of args )
    """

    def __init__(self, name='reactor'):
        self._workQueue = Queue()
        self._name = name
        self._funcsByRequest = {}
        self.running = True
        t = Thread(name=name, target=self._accept, daemon=True)
        t.start()

    def _accept(self):
        """
        Work loop runs forever (or until running is False)
        :return:
        """
        logger.warning("Reactor " + self._name + " is starting")
        while self.running:
            try:
                self._completeTask()
            except:
                logger.exception("Unexpected exception during request processing")
        logger.warning("Reactor " + self._name + " is terminating")

    def _completeTask(self):
        req = self._workQueue.get()
        if req is not None:
            try:
                logger.debug("Processing " + req[0])
                func = self._funcsByRequest.get(req[0])
                if func is not None:
                    func(*req[1])
                else:
                    logger.error("Ignoring unknown request on reactor " + self._name + " " + req[0])
            finally:
                self._workQueue.task_done()

    def register(self, requestType, function):
        """
        Registers a new function to handle the given request type.
        :param requestType:
        :param function:
        :return:
        """
        self._funcsByRequest[requestType] = function

    def offer(self, requestType, *args):
        """
        public interface to the reactor.
        :param requestType:
        :param args:
        :return:
        """
        if self._funcsByRequest.get(requestType) is not None:
            self._workQueue.put((requestType, list(*args)))
        else:
            logger.error("Ignoring unknown request on reactor " + self._name + " " + requestType)



from collections import OrderedDict

from core.BaseConfig import BaseConfig
from core.handler import CSVLogger, HttpPoster
from .i2c_io import WhiteNoiseProvider, mock_io
from .mpu6050 import mpu6050
from .smbus_io import smbus_io


class Config(BaseConfig):
    def __init__(self):
        super().__init__('recorder', defaultPort=10002)
        self.handlers = self._loadHandlers()
        self.recordingDevices = self._loadRecordingDevices()
        self.measurements = self._loadMeasurements()

    def loadDefaultConfig(self):
        return {
            'host': self.getDefaultHostname(),
            'port': 10002,
            'debug': False,
            'debugLogging': False,
            'useAsyncHandler': True,
            'accelerometers': [
                {
                    'name': 'mpu6050',
                    'type': 'mpu6050',
                    'fs': 500,
                    'handler': 'local log',
                    'io': {
                        'type': 'smbus',
                        'busId': 1
                    }
                }
            ],
            'handlers': [
                {
                    'name': 'remote',
                    'type': 'post',
                    'target': 'http://127.0.0.1:10001'
                }
            ]
        }

    def useAsyncHandlers(self):
        """
        :return: if async handling is on, defaults to True.
        """
        return self.config.get('useAsyncHandler', True)

    def getPingInterval(self):
        """
        :return: the server ping interval if set, defaults to 5s.
        """
        return self.config.get('pingInterval', 5)

    def createDevice(self, deviceCfg):
        """
        Creates a measurement deviceCfg from the input configuration.
        :param: deviceCfg: the deviceCfg cfg.
        :param: handlers: the loaded handlers.
        :return: the constructed deviceCfg.
        """
        ioCfg = deviceCfg['io']
        type = deviceCfg['type']
        if type == 'mpu6050':
            fs = deviceCfg.get('fs')
            if ioCfg['type'] == 'mock':
                provider = ioCfg.get('provider')
                if provider is not None and provider == 'white noise':
                    dataProvider = WhiteNoiseProvider()
                else:
                    raise ValueError(provider + " is not a supported mock io data provider")
                self.logger.warning("Loading mock data provider for mpu6050")
                io = mock_io(dataProvider=dataProvider.provide)
            elif ioCfg['type'] == 'smbus':
                busId = ioCfg['busId']
                self.logger.warning("Loading smbus %d", busId)
                io = smbus_io(busId)
            else:
                raise ValueError(ioCfg['type'] + " is not a supported io provider")

            return mpu6050(io, fs=fs)
        else:
            raise ValueError(type + " is not a supported device")

    def _loadRecordingDevices(self):
        """
        Loads the recordingDevices specified in the configuration.
        :param: handlers the loaded handlers.
        :return: the constructed recordingDevices in a dict keyed by name.
        """
        return {device.name: device for device in
                [self.createDevice(deviceCfg) for deviceCfg in self.config['accelerometers']]}

    def _loadMeasurements(self):
        """
        creates a dictionary to store a list of measurements by device.
        :param devices: the recordingDevices we have available.
        :return: a dictionary of device name -> {recording name -> measurements}
        """
        return {name: OrderedDict() for name in self.recordingDevices}

    def createHandler(self, handler):
        """
        Creates a data handler from the input configuration.
        :param handler: the handler cfg.
        :return: the constructed handler.
        """
        target = handler['target']
        if handler['type'] == 'log':
            self.logger.warning("Initialising csvlogger to log data to " + target)
            return CSVLogger('recorder', handler['name'], target)
        elif handler['type'] == 'post':
            self.logger.warning("Initialising http logger to log data to " + target)
            return HttpPoster(handler['name'], target)

    def _loadHandlers(self):
        """
        creates a dictionary of named handler instances
        :return: the dictionary
        """
        return {handler.name: handler for handler in map(self.createHandler, self.config['handlers'])}



import logging
import threading
from datetime import datetime
from enum import Enum

from flask import request
from flask_restful import Resource, marshal_with, fields

from core.interface import DATETIME_FORMAT, RecordingDeviceStatus

scheduledMeasurementFields = {
    'name': fields.String,
    'device': fields.String(attribute=lambda x: x.device.name if x is not None else None),
    'statuses': fields.Nested({'name': fields.String, 'time': fields.DateTime})
}

logger = logging.getLogger('recorder.measurements')


class Measurements(Resource):
    def __init__(self, **kwargs):
        self.measurements = kwargs['measurements']

    @marshal_with(scheduledMeasurementFields)
    def get(self, deviceId):
        """
        lists all known active measurements.
        """
        measurementsByName = self.measurements.get(deviceId)
        if measurementsByName is None:
            return []
        else:
            return list(measurementsByName.values())


class Measurement(Resource):
    def __init__(self, **kwargs):
        self.measurements = kwargs['measurements']
        self.recordingDevices = kwargs['recordingDevices']

    @marshal_with(scheduledMeasurementFields)
    def get(self, deviceId, measurementId):
        """
        details the specific measurement.
        """
        record = self.measurements.get(deviceId)
        if record is not None:
            return record.get(measurementId)
        return None

    @marshal_with(scheduledMeasurementFields)
    def put(self, deviceId, measurementId):
        """
        Schedules a new measurement at the specified time.
        :param deviceId: the device to measure.
        :param measurementId: the name of the measurement.
        :return: 200 if it was scheduled, 400 if the device is busy, 500 if the device is bad.
        """
        record = self.measurements.get(deviceId)
        if record is not None:
            measurement = record.get(measurementId)
            if measurement is not None:
                if len([x.name for x in measurement.statuses if x.name is 'COMPLETE' or x.name is 'FAILED']) > 0:
                    logger.info('Overwriting existing completed measurement ' + x.name)
                    measurement = None
            if measurement is None:
                logger.info('Initiating measurement ' + measurementId)
                measurement = ScheduledMeasurement(measurementId, self.recordingDevices.get(deviceId))
                body = request.get_json()
                duration_ = body['duration']
                def _cleanup():
                    logger.info('Removing ' + measurementId + ' from ' + deviceId)
                    record.pop(measurementId)
                measurement.schedule(duration_, at=body.get('at'), delay=body.get('delay'), callback=_cleanup)
                # a quick hack to enable the measurement to be cleaned up by the ScheduledMeasurement
                record[measurementId] = measurement
                return measurement, 200
            else:
                return measurement, 400
        else:
            return 'unknown device ' + deviceId, 400

    @marshal_with(scheduledMeasurementFields)
    def delete(self, deviceId, measurementId):
        """
        Deletes a stored measurement.
        :param deviceId: the device to measure.
        :param measurementId: the name of the measurement.
        :return: 200 if it was deleted, 400 if no such measurement (or device).
        """
        record = self.measurements.get(deviceId)
        if record is not None:
            popped = record.pop(measurementId, None)
            return popped, 200 if popped else 400
        return None, 400


class AbortMeasurement(Resource):
    def __init__(self, **kwargs):
        self.measurements = kwargs['measurements']
        self.recordingDevices = kwargs['recordingDevices']

    @marshal_with(scheduledMeasurementFields)
    def get(self, deviceId, measurementId):
        """
        signals a stop for the given measurement.
        :param deviceId: the device to measure.
        :param measurementId: the name of the measurement.
        :return: 200 if stop is signalled, 400 if it doesn't exist or is not running.
        """
        record = self.measurements.get(deviceId)
        if record is not None:
            measurement = record.get(measurementId)
            if measurement.recording:
                device = self.recordingDevices.get(deviceId)
                device.signalStop()
                return measurement, 200
            else:
                return measurement, 400
        return '', 400


class ScheduledMeasurementStatus(Enum):
    INITIALISING = 0
    SCHEDULED = 1
    RUNNING = 2
    COMPLETE = 3
    FAILED = 4


class ScheduledMeasurement:
    """
    A named measurement that is scheduled to run at a given time for a given duration.
    """

    def __init__(self, name, device):
        """
        create a new measurement with the given name to execute on the given device.
        :param name: the measurement name.
        :param device: the device to run on.
        """
        self.name = name
        self.device = device
        self.recording = False
        self.statuses = [{'name': ScheduledMeasurementStatus.INITIALISING.name, 'time': datetime.now()}]
        self.callback = None

    def schedule(self, duration, at=None, delay=None, callback=None):
        """
        schedules the measurement (to execute asynchronously).
        :param duration: how long to run for.
        :param at: the time to start at.
        :param delay: the time to wait til starting (use at or delay).
        :param callback: a callback.
        :return: nothing.
        """
        delay = self.calculateDelay(at, delay)
        self.callback = callback
        logger.info('Initiating measurement ' + self.name + ' for ' + str(duration) + 's in ' + str(delay) + 's')
        self.statuses.append({'name': ScheduledMeasurementStatus.SCHEDULED.name, 'time': datetime.now()})
        threading.Timer(delay, self.execute, [duration]).start()

    def execute(self, duration):
        """
        Executes the measurement, recording the event status.
        :param duration: the time to run for.
        :return: nothing.
        """
        self.statuses.append({'name': ScheduledMeasurementStatus.RUNNING.name, 'time': datetime.now()})
        try:
            self.recording = True
            self.device.start(self.name, durationInSeconds=duration)
        finally:
            self.recording = False
        if self.device.status == RecordingDeviceStatus.FAILED:
            self.statuses.append({'name': ScheduledMeasurementStatus.FAILED.name,
                                  'time': datetime.now(),
                                  'reason': self.device.failureCode})
        else:
            self.statuses.append({'name': ScheduledMeasurementStatus.COMPLETE.name, 'time': datetime.now()})
        # this is a bit of a hack, need to remove this at some point by refactoring the way measurements are stored
        if self.callback is not None:
            self.callback()

    def calculateDelay(self, at, delay):
        """
        Creates the delay from now til the specified start time, uses "at" if available.
        :param at: the start time in %a %b %d %H:%M:%S %Y format.
        :param delay: the delay from now til start.
        :return: the delay.
        """
        if at is not None:
            return max((datetime.strptime(at, DATETIME_FORMAT) - datetime.now()).total_seconds(), 0)
        elif delay is not None:
            return delay
        else:
            return 0



from enum import Enum

from flask_restful import fields
from flask_restful.fields import Raw, MarshallingException


class RecordingDeviceStatus(Enum):
    """
    the status of the recording device.
    """
    NEW = 0
    INITIALISED = 1
    RECORDING = 2
    FAILED = 3


class EnumField(Raw):
    """
    Marshal an enum as a string
    """

    def format(self, value):
        try:
            return value.name
        except ValueError as ve:
            raise MarshallingException(ve)


"""
Formats a TargetState to JSON
"""
targetStateFields = {
    'fs': fields.Integer,
    'samplesPerBatch': fields.Integer,
    'accelerometerEnabled': fields.Boolean,
    'accelerometerSens': fields.Integer,
    'gyroEnabled': fields.Boolean,
    'gyroSens': fields.Integer
}

"""
Formats a RecordingDevice (which is an Accelerometer under the covers) to JSON. This is targetStateFields (with some
custom attribute mapping) + device status.
"""
recordingDeviceFields = {
    # target state fields
    'name': fields.String,
    'fs': fields.Integer,
    'samplesPerBatch': fields.Integer,
    'accelerometerEnabled': fields.Boolean(attribute='_accelEnabled'),
    'accelerometerSens': fields.Integer(attribute='accelerometerSensitivity'),
    'gyroEnabled': fields.Boolean(attribute='_gyroEnabled'),
    'gyroSens': fields.Integer(attribute='gyroSensitivity'),
    # device status fields
    'status': EnumField,
    'failureCode': fields.String
}

DATETIME_FORMAT = '%Y%m%d_%H%M%S'
API_PREFIX = '/api/1'


import abc

import requests

class RequestException(Exception):
    """
    exists to decouple the caller from the requests specific exception
    """
    pass

class HttpClient(object):
    """
    A simple http client to be used by the various parts of the system to send data via http. Exists to facilitate
    easier unit testing than monkey patching the transport layer or faffing around with the requests mock.
    """

    @abc.abstractmethod
    def get(self, url, **kwargs):
        """
        sends a GET.
        :param url:  the url to send to.
        :param kwargs: the args.
        :return:
        """
        pass

    @abc.abstractmethod
    def put(self, url, **kwargs):
        """
        sends a PUT.
        :param url:  the url to send to.
        :param kwargs: the args.
        :return:
        """
        pass

    @abc.abstractmethod
    def patch(self, url, **kwargs):
        """
        sends a patch.
        :param url:  the url to send to.
        :param kwargs: the args.
        :return:
        """
        pass

    @abc.abstractmethod
    def post(self, url, **kwargs):
        """
        sends a post.
        :param url:  the url to send to.
        :param kwargs: the args.
        :return:
        """
        pass

    @abc.abstractmethod
    def delete(self, url, **kwargs):
        """
        sends a delete.
        :param url:  the url to send to.
        :param kwargs: the args.
        :return:
        """
        pass


class RequestsBasedHttpClient(HttpClient):
    """
    An implementation of HttpClient which simply delegates to the requests lib.
    """

    def get(self, url, **kwargs):
        try:
            return requests.get(url, timeout=0.5, **kwargs)
        except requests.exceptions.RequestException as e:
            raise RequestException("Unable to GET from " + url) from e

    def post(self, url, **kwargs):
        try:
            return requests.post(url, timeout=0.5, **kwargs)
        except requests.exceptions.RequestException as e:
            raise RequestException("Unable to POST to " + url) from e

    def put(self, url, **kwargs):
        try:
            return requests.put(url, timeout=0.5, **kwargs)
        except requests.exceptions.RequestException as e:
            raise RequestException("Unable to PUT " + url) from e

    def patch(self, url, **kwargs):
        try:
            return requests.patch(url, timeout=0.5, **kwargs)
        except requests.exceptions.RequestException as e:
            raise RequestException("Unable to PATCH " + url) from e

    def delete(self, url, **kwargs):
        try:
            return requests.delete(url, timeout=0.5, **kwargs)
        except requests.exceptions.RequestException as e:
            raise RequestException("Unable to DELETE " + url) from e


class RecordingHttpClient(HttpClient):
    """
    An implementation of httpclient intended for use by test cases as it simply records each call in a list.
    """

    def __init__(self):
        self.record = []

    def post(self, url, **kwargs):
        self.record.append(('post', url, kwargs))
        return self._resp()

    def _resp(self):
        from unittest.mock import Mock
        mock = Mock()
        mock.status_code = 200
        return mock

    def patch(self, url, **kwargs):
        self.record.append(('patch', url, kwargs))
        return self._resp()

    def put(self, url, **kwargs):
        self.record.append(('put', url, kwargs))
        return self._resp()

    def delete(self, url, **kwargs):
        self.record.append(('delete', url, kwargs))
        return self._resp()

    def get(self, url, **kwargs):
        self.record.append(('get', url, kwargs))
        return self._resp()



import abc
from queue import Queue

from .mpu6050 import mpu6050


class i2c_io(object):
    """
    A thin wrapper on the smbus for reading and writing data. Exists to allow unit testing without a real device
    connected.
    """

    def __init__(self):
        pass

    """
    Writes data to the device.
    :param: i2cAddress: the address to write to.
    :param: register: the location to write to.
    :param: val: the value to write.
    """

    @abc.abstractmethod
    def write(self, i2cAddress, register, val):
        pass

    """
    Reads data from the device.
    :param: i2cAddress: the address to read from.
    :param: register: the register to read from.
    :return: the data read.
    """

    @abc.abstractmethod
    def read(self, i2cAddress, register):
        pass

    """
    Reads a block of data from the device.
    :param: i2cAddress: the address to read from.
    :param: register: the register to read from.
    :param: length: no of bytes to read.
    :return: the data read.
    """

    @abc.abstractmethod
    def readBlock(self, i2cAddress, register, length):
        pass


class mock_io(i2c_io):
    def __init__(self, dataProvider=None):
        super().__init__()
        self.valuesWritten = []
        self.dataProvider = dataProvider
        self.valsToRead = Queue()

    def write(self, i2cAddress, register, val):
        self.valuesWritten.append([i2cAddress, register, val])

    def readBlock(self, i2cAddress, register, length):
        if self.dataProvider is not None:
            ret = self.dataProvider(register, length)
            if ret is not None:
                return ret
        return self.valsToRead.get_nowait()

    def read(self, i2cAddress, register):
        if self.dataProvider is not None:
            ret = self.dataProvider(register)
            if ret is not None:
                return ret
        return self.valsToRead.get_nowait()


class MockIoDataProvider:
    @abc.abstractmethod
    def provide(self, register):
        pass


class WhiteNoiseProvider(MockIoDataProvider):
    """
    A mock io provider which yields white noise.
    """

    def __init__(self):
        import numpy as np

        self.samples = {
            'x': np.random.normal(0, 0.25, size=1000),
            'y': np.random.normal(0, 0.25, size=1000),
            'z': np.random.normal(0, 0.25, size=1000)
        }
        self.idx = 0

    def provide(self, register, length=None):
        if register is mpu6050.MPU6050_RA_INT_STATUS:
            return 0x01
        elif register is mpu6050.MPU6050_RA_FIFO_COUNTH:
            return [0b0000, 0b1100]
        elif register is mpu6050.MPU6050_RA_FIFO_R_W:
            bytes = bytearray()
            self.addValue(bytes, 'x')
            self.addValue(bytes, 'y')
            self.addValue(bytes, 'z')
            self.idx += 1
            self.addValue(bytes, 'x')
            self.addValue(bytes, 'y')
            self.addValue(bytes, 'z')
            return bytes
        else:
            if length is None:
                return 0b00000000
            else:
                return [x.to_bytes(1, 'big') for x in range(length)]

    def addValue(self, bytes, key):
        val = abs(int((self.samples[key][self.idx % 1000] * 32768)))
        try:
            b = bytearray(val.to_bytes(2, 'big'))
        except OverflowError:
            print("Value too big - " + str(val) + " - replacing with 0")
            val = 0
            b = bytearray(val.to_bytes(2, 'big'))
        bytes.extend(b)



import collections
import logging
import struct
from _ctypes import ArgumentError
from time import sleep, time

from .accelerometer import Accelerometer, ACCEL_X, ACCEL_Y, ACCEL_Z, GYRO_X, GYRO_Y, GYRO_Z, TEMP, SAMPLE_TIME

SENSOR_SCALE_FACTOR = 32768.0
DEFAULT_GYRO_SENSITIVITY = 500.0
DEFAULT_ACCELEROMETER_SENSITIVITY = 2.0
DEFAULT_TEMPERATURE_GAIN = 1.0 / 340.0
DEFAULT_TEMPERATURE_OFFSET = 36.53

logger = logging.getLogger('recorder.mpu6050')


class mpu6050(Accelerometer):
    """
    Controls the Invensense MPU-6050, code based on danjperron's https://github.com/danjperron/mpu6050TestInC
    """
    # Temperature in degrees C = (TEMP_OUT Register Value as a signed quantity)/340 + 36.53
    _temperatureGain = DEFAULT_TEMPERATURE_GAIN
    _temperatureOffset = DEFAULT_TEMPERATURE_OFFSET

    # converted from Jeff Rowberg code https://github.com/jrowberg/i2cdevlib/blob/master/Arduino/MPU6050/MPU6050.h
    MPU6050_ADDRESS = 0x68  # default I2C Address

    # bit masks for enabling the which sensors are written to the FIFO
    enableAccelerometerMask = 0b00001000
    enableGyroMask = 0b01110000
    enableTemperatureMask = 0b10000000

    # register definition
    MPU6050_RA_XG_OFFS_TC = 0x00  # [7] PWR_MODE, [6:1] XG_OFFS_TC, [0] OTP_BNK_VLD
    MPU6050_RA_YG_OFFS_TC = 0x01  # [7] PWR_MODE, [6:1] YG_OFFS_TC, [0] OTP_BNK_VLD
    MPU6050_RA_ZG_OFFS_TC = 0x02  # [7] PWR_MODE, [6:1] ZG_OFFS_TC, [0] OTP_BNK_VLD
    MPU6050_RA_X_FINE_GAIN = 0x03  # [7:0] X_FINE_GAIN
    MPU6050_RA_Y_FINE_GAIN = 0x04  # [7:0] Y_FINE_GAIN
    MPU6050_RA_Z_FINE_GAIN = 0x05  # [7:0] Z_FINE_GAIN
    MPU6050_RA_XA_OFFS_H = 0x06  # [15:0] XA_OFFS
    MPU6050_RA_XA_OFFS_L_TC = 0x07
    MPU6050_RA_YA_OFFS_H = 0x08  # [15:0] YA_OFFS
    MPU6050_RA_YA_OFFS_L_TC = 0x09
    MPU6050_RA_ZA_OFFS_H = 0x0A  # [15:0] ZA_OFFS
    MPU6050_RA_ZA_OFFS_L_TC = 0x0B
    MPU6050_RA_XG_OFFS_USRH = 0x13  # [15:0] XG_OFFS_USR
    MPU6050_RA_XG_OFFS_USRL = 0x14
    MPU6050_RA_YG_OFFS_USRH = 0x15  # [15:0] YG_OFFS_USR
    MPU6050_RA_YG_OFFS_USRL = 0x16
    MPU6050_RA_ZG_OFFS_USRH = 0x17  # [15:0] ZG_OFFS_USR
    MPU6050_RA_ZG_OFFS_USRL = 0x18
    MPU6050_RA_SMPLRT_DIV = 0x19
    MPU6050_RA_CONFIG = 0x1A
    MPU6050_RA_GYRO_CONFIG = 0x1B
    MPU6050_RA_ACCEL_CONFIG = 0x1C
    MPU6050_RA_FF_THR = 0x1D
    MPU6050_RA_FF_DUR = 0x1E
    MPU6050_RA_MOT_THR = 0x1F
    MPU6050_RA_MOT_DUR = 0x20
    MPU6050_RA_ZRMOT_THR = 0x21
    MPU6050_RA_ZRMOT_DUR = 0x22
    MPU6050_RA_FIFO_EN = 0x23
    MPU6050_RA_I2C_MST_CTRL = 0x24
    MPU6050_RA_I2C_SLV0_ADDR = 0x25
    MPU6050_RA_I2C_SLV0_REG = 0x26
    MPU6050_RA_I2C_SLV0_CTRL = 0x27
    MPU6050_RA_I2C_SLV1_ADDR = 0x28
    MPU6050_RA_I2C_SLV1_REG = 0x29
    MPU6050_RA_I2C_SLV1_CTRL = 0x2A
    MPU6050_RA_I2C_SLV2_ADDR = 0x2B
    MPU6050_RA_I2C_SLV2_REG = 0x2C
    MPU6050_RA_I2C_SLV2_CTRL = 0x2D
    MPU6050_RA_I2C_SLV3_ADDR = 0x2E
    MPU6050_RA_I2C_SLV3_REG = 0x2F
    MPU6050_RA_I2C_SLV3_CTRL = 0x30
    MPU6050_RA_I2C_SLV4_ADDR = 0x31
    MPU6050_RA_I2C_SLV4_REG = 0x32
    MPU6050_RA_I2C_SLV4_DO = 0x33
    MPU6050_RA_I2C_SLV4_CTRL = 0x34
    MPU6050_RA_I2C_SLV4_DI = 0x35
    MPU6050_RA_I2C_MST_STATUS = 0x36
    MPU6050_RA_INT_PIN_CFG = 0x37
    MPU6050_RA_INT_ENABLE = 0x38
    MPU6050_RA_DMP_INT_STATUS = 0x39
    MPU6050_RA_INT_STATUS = 0x3A
    MPU6050_RA_ACCEL_XOUT_H = 0x3B
    MPU6050_RA_ACCEL_XOUT_L = 0x3C
    MPU6050_RA_ACCEL_YOUT_H = 0x3D
    MPU6050_RA_ACCEL_YOUT_L = 0x3E
    MPU6050_RA_ACCEL_ZOUT_H = 0x3F
    MPU6050_RA_ACCEL_ZOUT_L = 0x40
    MPU6050_RA_TEMP_OUT_H = 0x41
    MPU6050_RA_TEMP_OUT_L = 0x42
    MPU6050_RA_GYRO_XOUT_H = 0x43
    MPU6050_RA_GYRO_XOUT_L = 0x44
    MPU6050_RA_GYRO_YOUT_H = 0x45
    MPU6050_RA_GYRO_YOUT_L = 0x46
    MPU6050_RA_GYRO_ZOUT_H = 0x47
    MPU6050_RA_GYRO_ZOUT_L = 0x48
    MPU6050_RA_EXT_SENS_DATA_00 = 0x49
    MPU6050_RA_EXT_SENS_DATA_01 = 0x4A
    MPU6050_RA_EXT_SENS_DATA_02 = 0x4B
    MPU6050_RA_EXT_SENS_DATA_03 = 0x4C
    MPU6050_RA_EXT_SENS_DATA_04 = 0x4D
    MPU6050_RA_EXT_SENS_DATA_05 = 0x4E
    MPU6050_RA_EXT_SENS_DATA_06 = 0x4F
    MPU6050_RA_EXT_SENS_DATA_07 = 0x50
    MPU6050_RA_EXT_SENS_DATA_08 = 0x51
    MPU6050_RA_EXT_SENS_DATA_09 = 0x52
    MPU6050_RA_EXT_SENS_DATA_10 = 0x53
    MPU6050_RA_EXT_SENS_DATA_11 = 0x54
    MPU6050_RA_EXT_SENS_DATA_12 = 0x55
    MPU6050_RA_EXT_SENS_DATA_13 = 0x56
    MPU6050_RA_EXT_SENS_DATA_14 = 0x57
    MPU6050_RA_EXT_SENS_DATA_15 = 0x58
    MPU6050_RA_EXT_SENS_DATA_16 = 0x59
    MPU6050_RA_EXT_SENS_DATA_17 = 0x5A
    MPU6050_RA_EXT_SENS_DATA_18 = 0x5B
    MPU6050_RA_EXT_SENS_DATA_19 = 0x5C
    MPU6050_RA_EXT_SENS_DATA_20 = 0x5D
    MPU6050_RA_EXT_SENS_DATA_21 = 0x5E
    MPU6050_RA_EXT_SENS_DATA_22 = 0x5F
    MPU6050_RA_EXT_SENS_DATA_23 = 0x60
    MPU6050_RA_MOT_DETECT_STATUS = 0x61
    MPU6050_RA_I2C_SLV0_DO = 0x63
    MPU6050_RA_I2C_SLV1_DO = 0x64
    MPU6050_RA_I2C_SLV2_DO = 0x65
    MPU6050_RA_I2C_SLV3_DO = 0x66
    MPU6050_RA_I2C_MST_DELAY_CTRL = 0x67
    MPU6050_RA_SIGNAL_PATH_RESET = 0x68
    MPU6050_RA_MOT_DETECT_CTRL = 0x69
    MPU6050_RA_USER_CTRL = 0x6A
    MPU6050_RA_PWR_MGMT_1 = 0x6B
    MPU6050_RA_PWR_MGMT_2 = 0x6C
    MPU6050_RA_BANK_SEL = 0x6D
    MPU6050_RA_MEM_START_ADDR = 0x6E
    MPU6050_RA_MEM_R_W = 0x6F
    MPU6050_RA_DMP_CFG_1 = 0x70
    MPU6050_RA_DMP_CFG_2 = 0x71
    MPU6050_RA_FIFO_COUNTH = 0x72
    MPU6050_RA_FIFO_COUNTL = 0x73
    MPU6050_RA_FIFO_R_W = 0x74
    MPU6050_RA_WHO_AM_I = 0x75

    ZeroRegister = [
        MPU6050_RA_FF_THR,  # Freefall threshold of |0mg|  LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_FF_THR, 0x00);
        MPU6050_RA_FF_DUR,  # Freefall duration limit of 0   LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_FF_DUR, 0x00);
        MPU6050_RA_MOT_THR,  # Motion threshold of 0mg     LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_MOT_THR, 0x00);
        MPU6050_RA_MOT_DUR,  # Motion duration of 0s    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_MOT_DUR, 0x00);
        MPU6050_RA_ZRMOT_THR,  # Zero motion threshold    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_ZRMOT_THR, 0x00);
        MPU6050_RA_ZRMOT_DUR,
        # Zero motion duration threshold    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_ZRMOT_DUR, 0x00);
        MPU6050_RA_FIFO_EN,
        # Disable sensor output to FIFO buffer    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_FIFO_EN, 0x00);
        MPU6050_RA_I2C_MST_CTRL,
        # AUX I2C setup    //Sets AUX I2C to single master control, plus other config    LDByteWriteI2C(
        # MPU6050_ADDRESS, MPU6050_RA_I2C_MST_CTRL, 0x00);
        MPU6050_RA_I2C_SLV0_ADDR,
        # Setup AUX I2C slaves    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_ADDR, 0x00);
        MPU6050_RA_I2C_SLV0_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_REG, 0x00);
        MPU6050_RA_I2C_SLV0_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_CTRL, 0x00);
        MPU6050_RA_I2C_SLV1_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_ADDR, 0x00);
        MPU6050_RA_I2C_SLV1_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_REG, 0x00);
        MPU6050_RA_I2C_SLV1_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_CTRL, 0x00);
        MPU6050_RA_I2C_SLV2_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_ADDR, 0x00);
        MPU6050_RA_I2C_SLV2_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_REG, 0x00);
        MPU6050_RA_I2C_SLV2_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_CTRL, 0x00);
        MPU6050_RA_I2C_SLV3_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_ADDR, 0x00);
        MPU6050_RA_I2C_SLV3_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_REG, 0x00);
        MPU6050_RA_I2C_SLV3_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_CTRL, 0x00);
        MPU6050_RA_I2C_SLV4_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_ADDR, 0x00);
        MPU6050_RA_I2C_SLV4_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_REG, 0x00);
        MPU6050_RA_I2C_SLV4_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_DO, 0x00);
        MPU6050_RA_I2C_SLV4_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_CTRL, 0x00);
        MPU6050_RA_I2C_SLV4_DI,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_DI, 0x00);
        MPU6050_RA_INT_PIN_CFG,
        # MPU6050_RA_I2C_MST_STATUS //Read-only    //Setup INT pin and AUX I2C pass through    LDByteWriteI2C(
        # MPU6050_ADDRESS, MPU6050_RA_INT_PIN_CFG, 0x00);
        MPU6050_RA_INT_ENABLE,
        # Enable data ready interrupt      LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_INT_ENABLE, 0x00);
        MPU6050_RA_I2C_SLV0_DO,
        # Slave out, dont care    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_DO, 0x00);
        MPU6050_RA_I2C_SLV1_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_DO, 0x00);
        MPU6050_RA_I2C_SLV2_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_DO, 0x00);
        MPU6050_RA_I2C_SLV3_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_DO, 0x00);
        MPU6050_RA_I2C_MST_DELAY_CTRL,
        # More slave config      LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_MST_DELAY_CTRL, 0x00);
        MPU6050_RA_SIGNAL_PATH_RESET,
        # Reset sensor signal paths    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_SIGNAL_PATH_RESET, 0x00);
        MPU6050_RA_MOT_DETECT_CTRL,
        # Motion detection control    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_MOT_DETECT_CTRL, 0x00);
        MPU6050_RA_USER_CTRL,
        # Disables FIFO, AUX I2C, FIFO and I2C reset bits to 0    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_USER_CTRL,
        #  0x00);
        MPU6050_RA_CONFIG,  # Disable FSync, 256Hz DLPF    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_CONFIG, 0x00);
        MPU6050_RA_FF_THR,  # Freefall threshold of |0mg|    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_FF_THR, 0x00);
        MPU6050_RA_FF_DUR,  # Freefall duration limit of 0    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_FF_DUR, 0x00);
        MPU6050_RA_MOT_THR,  # Motion threshold of 0mg    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_MOT_THR, 0x00);
        MPU6050_RA_MOT_DUR,  # Motion duration of 0s    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_MOT_DUR, 0x00);
        MPU6050_RA_ZRMOT_THR,  # Zero motion threshold    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_ZRMOT_THR, 0x00);
        MPU6050_RA_ZRMOT_DUR,
        # Zero motion duration threshold    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_ZRMOT_DUR, 0x00);
        MPU6050_RA_FIFO_EN,
        # Disable sensor output to FIFO buffer    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_FIFO_EN, 0x00);
        MPU6050_RA_I2C_MST_CTRL,
        # AUX I2C setup    //Sets AUX I2C to single master control, plus other config    LDByteWriteI2C(
        # MPU6050_ADDRESS, MPU6050_RA_I2C_MST_CTRL, 0x00);
        MPU6050_RA_I2C_SLV0_ADDR,
        # Setup AUX I2C slaves    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_ADDR, 0x00);
        MPU6050_RA_I2C_SLV0_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_REG, 0x00);
        MPU6050_RA_I2C_SLV0_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_CTRL, 0x00);
        MPU6050_RA_I2C_SLV1_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_ADDR, 0x00);
        MPU6050_RA_I2C_SLV1_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_REG, 0x00);
        MPU6050_RA_I2C_SLV1_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_CTRL, 0x00);
        MPU6050_RA_I2C_SLV2_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_ADDR, 0x00);
        MPU6050_RA_I2C_SLV2_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_REG, 0x00);
        MPU6050_RA_I2C_SLV2_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_CTRL, 0x00);
        MPU6050_RA_I2C_SLV3_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_ADDR, 0x00);
        MPU6050_RA_I2C_SLV3_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_REG, 0x00);
        MPU6050_RA_I2C_SLV3_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_CTRL, 0x00);
        MPU6050_RA_I2C_SLV4_ADDR,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_ADDR, 0x00);
        MPU6050_RA_I2C_SLV4_REG,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_REG, 0x00);
        MPU6050_RA_I2C_SLV4_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_DO, 0x00);
        MPU6050_RA_I2C_SLV4_CTRL,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_CTRL, 0x00);
        MPU6050_RA_I2C_SLV4_DI,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV4_DI, 0x00);
        MPU6050_RA_I2C_SLV0_DO,
        # Slave out, dont care    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV0_DO, 0x00);
        MPU6050_RA_I2C_SLV1_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV1_DO, 0x00);
        MPU6050_RA_I2C_SLV2_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV2_DO, 0x00);
        MPU6050_RA_I2C_SLV3_DO,  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_SLV3_DO, 0x00);
        MPU6050_RA_I2C_MST_DELAY_CTRL,
        # More slave config    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_I2C_MST_DELAY_CTRL, 0x00);
        MPU6050_RA_SIGNAL_PATH_RESET,
        # Reset sensor signal paths    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_SIGNAL_PATH_RESET, 0x00);
        MPU6050_RA_MOT_DETECT_CTRL,
        # Motion detection control    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_MOT_DETECT_CTRL, 0x00);
        MPU6050_RA_USER_CTRL,
        # Disables FIFO, AUX I2C, FIFO and I2C reset bits to 0    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_USER_CTRL,
        #  0x00);
        MPU6050_RA_INT_PIN_CFG,
        # MPU6050_RA_I2C_MST_STATUS //Read-only    //Setup INT pin and AUX I2C pass through    LDByteWriteI2C(
        # MPU6050_ADDRESS, MPU6050_RA_INT_PIN_CFG, 0x00);
        MPU6050_RA_INT_ENABLE,
        # Enable data ready interrupt    LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_INT_ENABLE, 0x00);
        MPU6050_RA_FIFO_R_W]  # LDByteWriteI2C(MPU6050_ADDRESS, MPU6050_RA_FIFO_R_W, 0x00);

    def __init__(self, i2c_io, fs=None, samplesPerBatch=None, dataHandler=None):
        """
        initialises to a default state set to measure acceleration only.
        """
        super().__init__(fs, samplesPerBatch, dataHandler)
        self.name = 'mpu6050'
        self.fifoSensorMask = self.enableAccelerometerMask
        self._accelEnabled = True
        self._gyroEnabled = False
        self._setSampleSizeBytes()
        self.accelerometerSensitivity = DEFAULT_ACCELEROMETER_SENSITIVITY
        self.gyroSensitivity = DEFAULT_GYRO_SENSITIVITY
        self._accelerationFactor = self.accelerometerSensitivity / SENSOR_SCALE_FACTOR
        self._gyroFactor = self.gyroSensitivity / SENSOR_SCALE_FACTOR
        self.i2c_io = i2c_io
        self.doInit()

    def _setSampleSizeBytes(self):
        """
        updates the current record of the packet size per sample and the relationship between this and the fifo reads. 
        """
        self.sampleSizeBytes = self.getPacketSize()
        if self.sampleSizeBytes > 0:
            self.maxBytesPerFifoRead = (32 // self.sampleSizeBytes)

    def getPacketSize(self):
        """
        the current packet size.
        :return: the current packet size based on the enabled registers.
        """
        size = 0
        if self.isAccelerometerEnabled():
            size += 6
        if self.isGyroEnabled():
            size += 6
        if self.isTemperatureEnabled():
            size += 2
        return size

    def initialiseDevice(self):
        """
        performs initialisation of the device
        :param batchSize: the no of samples that each provideData call should yield
        :return:
        """
        logger.debug("Initialising device")
        self.getInterruptStatus()
        self.setAccelerometerSensitivity(self._accelerationFactor * 32768.0)
        self.setGyroSensitivity(self._gyroFactor * 32768.0)
        self.setSampleRate(self.fs)
        for loop in self.ZeroRegister:
            self.i2c_io.write(self.MPU6050_ADDRESS, loop, 0)
        # Sets clock source to gyro reference w/ PLL
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_PWR_MGMT_1, 0b00000010)
        # Controls frequency of wakeups in accel low power mode plus the sensor standby modes
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_PWR_MGMT_2, 0x00)
        # Enables any I2C master interrupt source to generate an interrupt
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_INT_ENABLE, 0x01)
        # enable the FIFO
        self.enableFifo()
        logger.debug("Initialised device")

    def isAccelerometerEnabled(self):
        """
        tests whether acceleration is currently enabled
        :return: true if it is
        """
        return (self.fifoSensorMask & self.enableAccelerometerMask) == self.enableAccelerometerMask

    def enableAccelerometer(self):
        """
        Specifies the device should write acceleration values to the FIFO, is not applied until enableFIFO is called.
        :return:
        """
        logger.debug("Enabling acceleration sensor")
        self.fifoSensorMask |= self.enableAccelerometerMask
        self._accelEnabled = True
        self._setSampleSizeBytes()

    def disableAccelerometer(self):
        """
        Specifies the device should NOT write acceleration values to the FIFO, is not applied until enableFIFO is
        called.
        :return: 
        """
        logger.debug("Disabling acceleration sensor")
        self.fifoSensorMask &= ~self.enableAccelerometerMask
        self._accelEnabled = False
        self._setSampleSizeBytes()

    def isGyroEnabled(self):
        """
        tests whether gyro is currently enabled
        :return: true if it is
        """
        return (self.fifoSensorMask & self.enableGyroMask) == self.enableGyroMask

    def enableGyro(self):
        """
        Specifies the device should write gyro values to the FIFO, is not applied until enableFIFO is called.
        :return: 
        """
        logger.debug("Enabling gyro sensor")
        self.fifoSensorMask |= self.enableGyroMask
        self._gyroEnabled = True
        self._setSampleSizeBytes()

    def disableGyro(self):
        """
        Specifies the device should NOT write gyro values to the FIFO, is not applied until enableFIFO is called.
        :return: 
        """
        logger.debug("Disabling gyro sensor")
        self.fifoSensorMask &= ~self.enableGyroMask
        self._gyroEnabled = False
        self._setSampleSizeBytes()

    def isTemperatureEnabled(self):
        """
        tests whether temperature is currently enabled
        :return: true if it is
        """
        return (self.fifoSensorMask & self.enableTemperatureMask) == self.enableTemperatureMask

    def enableTemperature(self):
        """
        Specifies the device should write temperature values to the FIFO, is not applied until enableFIFO is called.
        :return: 
        """
        logger.debug("Enabling temperature sensor")
        self.fifoSensorMask |= self.enableTemperatureMask
        self._setSampleSizeBytes()

    def disableTemperature(self):
        """
        Specifies the device should NOT write temperature values to the FIFO, is not applied until enableFIFO is called.
        :return: 
        """
        logger.debug("Disabling temperature sensor")
        self.fifoSensorMask &= ~self.enableTemperatureMask
        self._setSampleSizeBytes()

    def setGyroSensitivity(self, value):
        """
        Sets the gyro sensitivity to 250, 500, 1000 or 2000 according to the given value (and implicitly disables the
        self
        tests)
        :param value: the target sensitivity.
        """
        try:
            self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_GYRO_CONFIG,
                              {250: 0, 500: 8, 1000: 16, 2000: 24}[value])
            self._gyroFactor = value / 32768.0
            self.gyroSensitivity = value
            logger.debug("Set gyro sensitivity = %d", value)
        except KeyError:
            raise ArgumentError(value + " is not a valid sensitivity (250,500,1000,2000)")

    def setAccelerometerSensitivity(self, value):
        """
        Sets the accelerometer sensitivity to 2, 4, 8 or 16 according to the given value. Throws an ArgumentError if
        the value provided is not valid.
        :param value: the target sensitivity.
        """
        # note that this implicitly disables the self tests on each axis
        # i.e. the full byte is actually 000[accel]000 where the 1st 3 are the accelerometer self tests, the next two
        # values are the actual sensitivity and the last 3 are unused
        # the 2 [accel] bits are translated by the device as follows; 00 = 2g, 01 = 4g, 10 = 8g, 11 = 16g
        # in binary we get 2 = 0, 4 = 1000, 8 = 10000, 16 = 11000
        # so the 1st 3 bits are always 0
        try:
            self.i2c_io.write(self.MPU6050_ADDRESS,
                              self.MPU6050_RA_ACCEL_CONFIG,
                              {2: 0, 4: 8, 8: 16, 16: 24}[value])
            self._accelerationFactor = value / 32768.0
            self.accelerometerSensitivity = value
            logger.debug("Set accelerometer sensitivity = %d", value)
        except KeyError:
            raise ArgumentError(value + " is not a valid sensitivity (2,4,8,18)")

    def setSampleRate(self, targetSampleRate):
        """
        Sets the internal sample rate of the MPU-6050, this requires writing a value to the device to set the sample
        rate as Gyroscope Output Rate / (1 + SMPLRT_DIV) where the gryoscope outputs at 8kHz and the peak sampling rate
         is 1kHz. The target sample rate is therefore capped at 1kHz.
        :param targetSampleRate: the target sample rate.
        :return:
        """
        sampleRateDenominator = int((8000 / min(targetSampleRate, 1000)) - 1)
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_SMPLRT_DIV, sampleRateDenominator)
        self.fs = 8000.0 / (sampleRateDenominator + 1.0)
        logger.debug("Set sample rate = %d", self.fs)

    def resetFifo(self):
        """
        Resets the FIFO by first disabling the FIFO then sending a FIFO_RESET and then re-enabling the FIFO.
        :return:
        """
        logger.debug("Resetting FIFO")
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_USER_CTRL, 0b00000000)
        pass
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_USER_CTRL, 0b00000100)
        pass
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_USER_CTRL, 0b01000000)
        self.getInterruptStatus()

    def enableFifo(self):
        """
        Enables the FIFO, resets it and then sets which values should be written to the FIFO.
        :return:
        """
        logger.debug("Enabling FIFO")
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_FIFO_EN, 0)
        self.resetFifo()
        self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_FIFO_EN, self.fifoSensorMask)
        logger.debug("Enabled FIFO")

    def getInterruptStatus(self):
        """
         reads and clears the current interrupt status. The byte that can be read is
         0b00010000 = FIFO_OFLOW_INT = FIFO has overflowed
         0b00001000 = I2C_MST_INT = I2C master interrupt has fired
         0b00000001 = DATA_RDY_INT = data is available to read
        """
        return self.i2c_io.read(self.MPU6050_ADDRESS, self.MPU6050_RA_INT_STATUS)

    def getFifoCount(self):
        """
        gets the amount of data available on the FIFO right now.
        :return: the number of bytes available on the FIFO which will be proportional to the number of samples available
        based on the values the device is configured to sample.
        """
        bytes = self.i2c_io.readBlock(self.MPU6050_ADDRESS, self.MPU6050_RA_FIFO_COUNTH, 2)
        count = (bytes[0] << 8) + bytes[1]
        logger.debug("FIFO Count: %d", count)
        return count

    def getDataFromFIFO(self, bytesToRead):
        """
        reads the specified number of bytes from the FIFO, should be called after a call to getFifoCount to ensure there
        is new data available (to avoid reading duplicate data).
        :param bytesToRead: the number of bytes to read.
        :return: the bytes read.
        """
        return self.i2c_io.readBlock(self.MPU6050_ADDRESS, self.MPU6050_RA_FIFO_R_W, bytesToRead)

    def provideData(self):
        """
        reads a batchSize batch of data from the FIFO while attempting to optimise the number of times we have to read
        from the device itself.
        :return: a list of data where each item is a single sample of data converted into real values and stored as a
        dict.
        """
        samples = []
        fifoBytesAvailable = 0
        fifoWasReset = False
        logger.debug(">> provideData target %d samples", self.samplesPerBatch)
        iterations = 0
        # allow 1.5x the expected duration of the batch
        breakTime = time() + ((self.samplesPerBatch / self.fs) * 1.5)
        overdue = False
        while len(samples) < self.samplesPerBatch and not overdue:
            iterations += 1
            if iterations > self.samplesPerBatch and iterations % 100 == 0:
                if time() > breakTime:
                    logger.warning("Breaking measurement after %d iterations, batch overdue", iterations)
                    overdue = True
            if fifoBytesAvailable < self.sampleSizeBytes or fifoWasReset:
                interrupt = self.getInterruptStatus()
                fifoBytesAvailable = self.getFifoCount()
                fifoWasReset = False
            logger.debug("Start sample loop [available: %d , required: %d]", fifoBytesAvailable, self.sampleSizeBytes)
            if interrupt & 0x10:
                logger.error("FIFO OVERFLOW, RESETTING [available: %d , interrupt: %d]", fifoBytesAvailable, interrupt)
                self.measurementOverflowed = True
                self.resetFifo()
                fifoWasReset = True
            elif fifoBytesAvailable == 1024:
                logger.error("FIFO FULL, RESETTING [available: %d , interrupt: %d]", fifoBytesAvailable, interrupt)
                self.measurementOverflowed = True
                self.resetFifo()
                fifoWasReset = True
            elif interrupt & 0x02 or interrupt & 0x01:
                # wait for at least 1 sample to arrive, should be a VERY short wait
                while fifoBytesAvailable < self.sampleSizeBytes:
                    logger.debug("Waiting for sample [available: %d , required: %d]", fifoBytesAvailable,
                                 self.sampleSizeBytes)
                    fifoBytesAvailable = self.getFifoCount()
                logger.debug("Processing data [available: %d , required: %d]", fifoBytesAvailable, self.sampleSizeBytes)
                fifoReadBytes = self.sampleSizeBytes
                # TODO this chunk of code is a bit messy, tidy it up
                # if we have more than 1 sample available then ensure we read as many as we can at once (albeit within
                # the limits of the max i2c read size of 32 bytes)
                if fifoBytesAvailable > self.sampleSizeBytes:
                    fifoReadBytes = min(fifoBytesAvailable // self.sampleSizeBytes,
                                        self.maxBytesPerFifoRead) * self.sampleSizeBytes
                    logger.debug("Excess bytes to read [available: %d , reading: %d]", fifoBytesAvailable,
                                 fifoReadBytes)
                # but don't read more than we need to fulfil the batch
                samplesToRead = fifoReadBytes // self.sampleSizeBytes
                excessSamples = self.samplesPerBatch - len(samples) - samplesToRead
                if excessSamples < 0:
                    samplesToRead += excessSamples
                    fifoReadBytes = int(samplesToRead * self.sampleSizeBytes)
                    logger.debug("Excess samples to read [available: %d , reading: %d]", fifoBytesAvailable,
                                 fifoReadBytes)
                else:
                    logger.debug("Reading [available: %d , reading: %d]", fifoBytesAvailable, fifoReadBytes)
                # read the bytes from the fifo, break it into sample sized chunks and convert to the actual values
                fifoBytes = self.getDataFromFIFO(fifoReadBytes)
                samples.extend([self.unpackSample(fifoBytes[i:i + self.sampleSizeBytes])
                                for i in range(0, len(fifoBytes), self.sampleSizeBytes)])
                # track the count here so we can avoid going back to the FIFO each time
                fifoBytesAvailable -= fifoReadBytes
                logger.debug("End sample loop [available: %d , required: %d]", fifoBytesAvailable, self.sampleSizeBytes)
        logger.debug("<< provideData %d samples", len(samples))
        return samples

    def unpackSample(self, rawData):
        """
        unpacks a single sample of data (where sample length is based on the currently enabled sensors).
        :param rawData: the data to convert
        :return: a converted data set.
        """
        length = len(rawData)
        # TODO error if not multiple of 2
        # logger.debug(">> unpacking sample %d length %d", self._sampleIdx, length)
        unpacked = struct.unpack(">" + ('h' * (length // 2)), memoryview(bytearray(rawData)).tobytes())
        # store the data in a dictionary
        mpu6050 = collections.OrderedDict()
        mpu6050[SAMPLE_TIME] = self._sampleIdx / self.fs
        sensorIdx = 0
        if self.isAccelerometerEnabled():
            mpu6050[ACCEL_X] = unpacked[sensorIdx] * self._accelerationFactor
            sensorIdx += 1
            mpu6050[ACCEL_Y] = unpacked[sensorIdx] * self._accelerationFactor
            sensorIdx += 1
            mpu6050[ACCEL_Z] = unpacked[sensorIdx] * self._accelerationFactor
            sensorIdx += 1

        if self.isTemperatureEnabled():
            mpu6050[TEMP] = unpacked[sensorIdx] * self._temperatureGain + self._temperatureOffset
            sensorIdx += 1

        if self.isGyroEnabled():
            mpu6050[GYRO_X] = unpacked[sensorIdx] * self._gyroFactor
            sensorIdx += 1
            mpu6050[GYRO_Y] = unpacked[sensorIdx] * self._gyroFactor
            sensorIdx += 1
            mpu6050[GYRO_Z] = unpacked[sensorIdx] * self._gyroFactor
            sensorIdx += 1
        # TODO should we send as a dict so the keys are available?
        output = list(mpu6050.values())
        self._sampleIdx += 1
        # logger.debug("<< unpacked sample length %d into vals size %d", length, len(output))
        return output

    def performSelfTest(self):
        """
        performs a self calibration against factory settings.
        :return:
        boolean: true if all sensors pass self test
        dictionary: all underlying percentage deviations, +/-14% is a pass
        """
        try:
            logger.debug(">> selfTest")
            # enable self test on all axes and set range to +/-250
            self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_GYRO_CONFIG, 0b11100000)
            # enable self test on all axes and set range to +/-8g
            self.i2c_io.write(self.MPU6050_ADDRESS, self.MPU6050_RA_ACCEL_CONFIG, 0b11110000)
            # sleep to let the test run
            sleep(0.5)
            # read the resulting raw data
            rawData = [self.i2c_io.read(self.MPU6050_ADDRESS, 0x0D),
                       self.i2c_io.read(self.MPU6050_ADDRESS, 0x0E),
                       self.i2c_io.read(self.MPU6050_ADDRESS, 0x0F),
                       self.i2c_io.read(self.MPU6050_ADDRESS, 0x10)]
            # each value is a 5 bit unsigned int that packs in both gyro and acceleration results
            # acceleration is in bits 5-7
            acceleration = {
                'x': (rawData[0] >> 3) | (rawData[3] & 0x30) >> 4,
                'y': (rawData[1] >> 3) | (rawData[3] & 0x0C) >> 4,
                'z': (rawData[2] >> 3) | (rawData[3] & 0x03) >> 4
            }
            # calculate results
            factoryTrimAcceleration = {
                'x': (4096.0 * 0.34) * (pow((0.92 / 0.34), ((acceleration['x'] - 1.0) / 30.0))),
                'y': (4096.0 * 0.34) * (pow((0.92 / 0.34), ((acceleration['y'] - 1.0) / 30.0))),
                'z': (4096.0 * 0.34) * (pow((0.92 / 0.34), ((acceleration['z'] - 1.0) / 30.0)))
            }
            # gyro is in bits 0-4
            gyro = {
                'x': rawData[0] & 0x1F,
                'y': rawData[1] & 0x1F,
                'z': rawData[2] & 0x1F
            }
            # calculate results
            factoryTrimGyro = {
                'x': (25.0 * 131.0) * (pow(1.046, (gyro['x'] - 1.0))),
                'y': (-25.0 * 131.0) * (pow(1.046, (gyro['y'] - 1.0))),
                'z': (25.0 * 131.0) * (pow(1.046, (gyro['z'] - 1.0)))
            }
            # percent diffs
            results = {}
            for axis in ['x', 'y', 'z']:
                results['a' + axis] = 100.0 + 100.0 * (acceleration[axis] - factoryTrimAcceleration[axis]) / \
                                              factoryTrimAcceleration[axis]
                results['g' + axis] = 100.0 + 100.0 * (gyro[axis] - factoryTrimGyro[axis]) / factoryTrimGyro[axis]

            passed = all(abs(v) < 14 for v in results.values())
            if passed:
                logger.debug("self test passed")
            else:
                logger.error("SELF TEST FAILURE " + str(results))
            return passed, results
        finally:
            self.setAccelerometerSensitivity(self.accelerometerSensitivity)
            self.setGyroSensitivity(self.gyroSensitivity)
            logger.debug("<< selfTest")



import abc
import csv
import datetime
import logging
import os
import threading
from queue import Queue, Empty

from flask import json

from core.httpclient import RequestsBasedHttpClient
from core.interface import DATETIME_FORMAT, API_PREFIX


class DataHandler:
    """
    A simple interface to define the expected behaviour of something that handles data.
    """

    @abc.abstractmethod
    def start(self, measurementId):
        """
        Initialises a handling session.
        :param measurementId: the name of the measurement that is about to start.
        :param converter: an optional converter to apply before the raw data is handled.
        :return:
        """
        pass

    @abc.abstractmethod
    def handle(self, data):
        """
        A callback for handling some raw data.
        :param data a list of dicts containing the sample data
        :return:
        """
        pass

    @abc.abstractmethod
    def stop(self, measurementId, failureReason=None):
        """
        Allows for cleanup on end of measurement.
        :param measurementId the measurement that is stopping.
        :param failureReason the reason it failed if any, if None then it completed ok.
        :return:
        """
        pass


class Discarder(DataHandler):
    """
    a data handler that simply throws the data away
    """

    def start(self, measurementId):
        pass

    def handle(self, data):
        pass

    def stop(self, measurementId, failureReason=None):
        pass


class CSVLogger(DataHandler):
    """
    A handler which logs the received data to CSV into target/measurementId/loggerName/data.out
    """

    def __init__(self, owner, name, target):
        self.logger = logging.getLogger(owner + '.csvlogger')
        self.name = name
        self.target = target
        self.started = False
        self._csv = None
        self._csvfile = None
        self._first = True

    def start(self, measurementId):
        targetDir = os.path.join(self.target, measurementId, self.name)
        if not os.path.exists(targetDir):
            os.makedirs(targetDir, exist_ok=True)
        targetPath = os.path.join(targetDir, 'data.out')
        if os.path.exists(targetPath):
            mode = 'w'
        else:
            mode = 'x'
        self._csvfile = open(targetPath, mode=mode, newline='\n')
        self._csv = csv.writer(self._csvfile)
        self.started = True
        self._first = True

    def handle(self, data):
        """
        Writes each sample to a csv file.
        :param data: the samples.
        :return:
        """
        self.logger.debug("Handling " + str(len(data)) + " data items")
        for datum in data:
            if isinstance(datum, dict):
                # these have to wrapped in a list for python 3.4 due to a change in the implementation
                # of OrderedDict in python 3.5+ (which means .keys() and .values() are sequences in 3.5+)
                if self._first:
                    self._csv.writerow(list(datum.keys()))
                    self._first = False
                self._csv.writerow(list(datum.values()))
            elif isinstance(datum, list):
                self._csv.writerow(datum)
            else:
                self.logger.warning("Ignoring unsupported data type " + str(type(datum)) + " : " + str(datum))

    def stop(self, measurementId, failureReason=None):
        if self._csvfile is not None:
            self.logger.debug("Closing csvfile for " + measurementId)
            self._csvfile.close()


class AsyncHandler(DataHandler):
    """
    A handler which hands the data off to another thread.
    """

    def __init__(self, owner, delegate):
        self.logger = logging.getLogger(owner + '.asynchandler')
        self.name = "Async"
        self.delegate = delegate
        self.queue = Queue()
        self.worker = None
        self.working = False
        self.stopping = False

    def start(self, measurementId):
        self.delegate.start(measurementId)
        self.worker = threading.Thread(target=self.asyncHandle, daemon=True)
        self.working = True
        self.logger.info('Starting async handler for ' + measurementId)
        self.worker.start()

    def handle(self, data):
        self.queue.put(data)

    def stop(self, measurementId, failureReason=None):
        # TODO do we need to link this stop to the status of the accelerometer
        self.logger.info('Stopping async handler for ' + measurementId)
        self.stopping = True
        self.queue.join()
        self.delegate.stop(measurementId, failureReason=failureReason)
        self.logger.info('Stopped async handler for ' + measurementId)
        self.working = False

    def asyncHandle(self):
        remaining = -1
        while self.working:
            try:
                event = self.queue.get(timeout=1)
                if event is not None:
                    self.delegate.handle(event)
                    self.queue.task_done()
                    if self.logger.isEnabledFor(logging.DEBUG):
                        self.logger.debug('async queue has ' + str(self.queue.qsize()) + ' items')
                    elif self.stopping:
                        if remaining == -1:
                            remaining = self.queue.qsize()
                        self.logger.info('Closing down asynchandler, ' + str(remaining) + ' items remaining')
                        remaining -= 1
            except Empty:
                pass


class HttpPoster(DataHandler):
    """
    A handler which sends the data over http.
    """

    def __init__(self, name, target, httpclient=RequestsBasedHttpClient()):
        self.name = name
        self.logger = logging.getLogger(name + '.httpposter')
        self.httpclient = httpclient
        self.target = target[:-1] if target.endswith('/') else target
        self.deviceName = None
        self.rootURL = self.target + API_PREFIX + '/measurements/'
        self.sendURL = None
        self.startResponseCode = None
        self.dataResponseCode = []
        self.endResponseCode = None

    def start(self, measurementId):
        """
        Posts to the target to tell it a named measurement is starting.
        :param measurementId:
        """
        self.sendURL = self.rootURL + measurementId + '/' + self.deviceName
        self.startResponseCode = self._doPut(self.sendURL)

    def _doPut(self, url, data=None):
        formattedPayload = None if data is None else json.dumps(data, sort_keys=True)
        try:
            return self.httpclient.put(url, json=formattedPayload).status_code
        except Exception as e:
            self.logger.exception(e)
            return 500

    def handle(self, data):
        """
        puts the data in the target.
        :param data: the data to post.
        :return:
        """
        self.dataResponseCode.append(self._doPut(self.sendURL + '/data', data=data))

    def stop(self, measurementId, failureReason=None):
        """
        informs the target the named measurement has completed
        :param measurementId: the measurement that has completed.
        :return:
        """
        if failureReason is None:
            self.endResponseCode = self._doPut(self.sendURL + "/complete")
        else:
            self.endResponseCode = self._doPut(self.sendURL + "/failed", data={'failureReason': failureReason})
        self.sendURL = None
        # TODO verify that the response codes are all ok



import abc
import logging
import os
from logging import handlers
from os import environ
from os import path

import yaml


class BaseConfig(object):
    def __init__(self, name, defaultPort=10001):
        self._name = name
        self.logger = logging.getLogger(name + '.config')
        self.config = self._loadConfig()
        self._hostname = self.config.get('host', self.getDefaultHostname())
        self._port = self.config.get('port', defaultPort)
        self._serviceURL = 'http://' + self.getHostname() + ':' + str(self.getPort())

    def getDefaultHostname(self):
        import socket
        return socket.getfqdn()

    def runInDebug(self):
        """
        :return: if debug mode is on, defaults to False.
        """
        return self.config.get('debug', False)

    def isDebugLogging(self):
        """
        :return: if debug logging mode is on, defaults to False.
        """
        return self.config.get('debugLogging', False)

    def getHostname(self):
        """
        :return: the host the device is running on, defaults to that found by a call to socket.getfqdn()
        """
        return self._hostname

    def getPort(self):
        """
        :return: the port to listen on, defaults to 10001
        """
        return self._port

    def getServiceURL(self):
        """
        :return: the address on which this service is listening.
        """
        return self._serviceURL

    def _loadConfig(self):
        """
        loads configuration from some predictable locations.
        :return: the config.
        """
        configPath = path.join(self._getConfigPath(), self._name + ".yml")
        if os.path.exists(configPath):
            self.logger.warning("Loading config from " + configPath)
            with open(configPath, 'r') as yml:
                return yaml.load(yml)
        defaultConfig = self.loadDefaultConfig()
        self._storeConfig(defaultConfig, configPath)
        return defaultConfig

    def _storeConfig(self, config, configPath):
        """ 
        Writes the config to the configPath.
        :param config a dict of config.
        :param configPath the path to the file to write to, intermediate dirs will be created as necessary.
        """
        self.logger.info("Writing to " + str(configPath))
        os.makedirs(os.path.dirname(configPath), exist_ok=True)
        with (open(configPath, 'w')) as yml:
            yaml.dump(config, yml, default_flow_style=False)

    @abc.abstractmethod
    def loadDefaultConfig(self):
        """
        Creates a default config bundle.
        :return:
        """
        pass

    def _getConfigPath(self):
        """
        Gets the currently configured config path.
        :return: the path, raises ValueError if it doesn't exist.
        """
        confHome = environ.get('VIBE_CONFIG_HOME')
        return confHome if confHome is not None else path.join(path.expanduser("~"), '.vibe')

    def configureLogger(self):
        """
        Configures the python logging system to log to a debug file and to stdout for warn and above.
        :return:
        """
        baseLogLevel = logging.DEBUG if self.isDebugLogging() else logging.INFO
        # create recorder app root logger
        logger = logging.getLogger(self._name)
        logger.setLevel(baseLogLevel)
        # file handler
        fh = handlers.RotatingFileHandler(path.join(self._getConfigPath(), self._name + '.log'),
                                          maxBytes=10 * 1024 * 1024, backupCount=10)
        fh.setLevel(baseLogLevel)
        # create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.WARN)
        # create formatter and add it to the handlers
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s')
        fh.setFormatter(formatter)
        ch.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(fh)
        logger.addHandler(ch)



from smbus2 import SMBus

from .i2c_io import i2c_io


class smbus_io(i2c_io):
    """
    an implementation of i2c_io which talks over the smbus.
    """

    def __init__(self, busId=1):
        super().__init__()
        self.bus = SMBus(bus=busId)

    """
    Delegates to smbus.write_byte_data
    """

    def write(self, i2cAddress, register, val):
        return self.bus.write_byte_data(i2cAddress, register, val)

    """
    Delegates to smbus.read_byte_data
    """

    def read(self, i2cAddress, register):
        return self.bus.read_byte_data(i2cAddress, register)

    """
    Delegates to smbus.read_i2c_block_data
    """

    def readBlock(self, i2cAddress, register, length):
        return self.bus.read_i2c_block_data(i2cAddress, register, length)



# a hack to work around fctnl when running on windows, for local dev purposes only
def fcntl(fd, op, arg=0):
    return 0


def ioctl(fd, op, arg=0, mutable_flag=True):
    if mutable_flag:
        return 0
    else:
        return ""


def flock(fd, op):
    return


def lockf(fd, operation, length=0, start=0, whence=0):
    return



import logging

from flask import request
from flask_restful import Resource, marshal_with

from core.interface import recordingDeviceFields
from ..common.accelerometer import RecordingDeviceStatus

logger = logging.getLogger('recorder.recordingDevices')


class RecordingDevices(Resource):
    def __init__(self, **kwargs):
        self.recordingDevices = kwargs['recordingDevices']

    @marshal_with(recordingDeviceFields)
    def get(self):
        """ lists all known recordingDevices. """
        return list(self.recordingDevices.values())


class RecordingDevice(Resource):
    def __init__(self, **kwargs):
        self.recordingDevices = kwargs['recordingDevices']
        self.heartbeater = kwargs['heartbeater']

    @marshal_with(recordingDeviceFields)
    def get(self, deviceId):
        """
        provides details on the specific device.
        :param: deviceId the device id.
        """
        return self.recordingDevices.get(deviceId)

    @marshal_with(recordingDeviceFields)
    def patch(self, deviceId):
        """
        Updates the device with the given data. Supports a json payload like
        {
            fs: newFs
            samplesPerBatch: samplesPerBatch
            gyroEnabled: true
            gyroSensitivity: 500
            accelerometerEnabled: true
            accelerometerSensitivity: 2
        }
        A heartbeat is sent on completion of the request to ensure the analyser gets a rapid update.
        :return: the device and 200 if the update was ok, 400 if not.
        """
        try:
            device = self.recordingDevices.get(deviceId)
            if device.status == RecordingDeviceStatus.INITIALISED:
                errors = self._handlePatch(device)
                if len(errors) == 0:
                    return device, 200
                else:
                    return device, 500
            else:
                return device, 400
        finally:
            logger.info("Sending adhoc heartbeat on device state update")
            self.heartbeater.sendHeartbeat()

    def _handlePatch(self, device):
        errors = []
        body = request.get_json()
        newFs = body.get('fs')
        if newFs is not None:
            if newFs != device.fs:
                device.setSampleRate(newFs)

        newSamplesPerBatch = body.get('samplesPerBatch')
        if newSamplesPerBatch is not None:
            if newSamplesPerBatch != device.samplesPerBatch:
                device.samplesPerBatch = newSamplesPerBatch

        enabled = body.get('gyroEnabled')
        if enabled and not device.isGyroEnabled():
            device.enableGyro()
        elif device.isGyroEnabled and not enabled:
            device.disableGyro()

        sensitivity = body.get('gyroSens')
        if sensitivity is not None:
            if sensitivity != device.gyroSensitivity:
                try:
                    device.setGyroSensitivity(sensitivity)
                except:
                    logger.exception("Invalid gyro sensitivity " + sensitivity)
                    errors.append("gyro sensitivity " + sensitivity)

        enabled = body.get('accelerometerEnabled')
        if enabled and not device.isAccelerometerEnabled():
            device.enableAccelerometer()
        elif device.isAccelerometerEnabled and not enabled:
            device.disableAccelerometer()

        sensitivity = body.get('accelerometerSens')
        if sensitivity is not None:
            if sensitivity != device.accelerometerSensitivity:
                try:
                    device.setAccelerometerSensitivity(sensitivity)
                except:
                    logger.exception("Invalid accelerometer sensitivity " + sensitivity)
                    errors.append("accelerometer sensitivity " + sensitivity)

        return errors


class SelfTest(Resource):
    def __init__(self, **kwargs):
        self.recordingDevices = kwargs['recordingDevices']

    def get(self, deviceId):
        """
        performs a self test on the given device.
        :param: deviceId the device id.
        :return: the self test results and 200 if it passed, 500 if it didn't
        """
        device = self.recordingDevices.get(deviceId)
        passed, results = device.performSelfTest()
        return results, 200 if passed else 500



import logging

from flask_restful import marshal

from core.interface import recordingDeviceFields, API_PREFIX

logger = logging.getLogger('recorder.heartbeater')


class Heartbeater(object):
    def __init__(self, httpclient, cfg, serverURL=None):
        self.httpclient = httpclient
        self.cfg = cfg
        self.serverURL = serverURL

    def ping(self):
        """
        Posts the current state of each device to the server and schedules the next call in n seconds.
        :param serverURL:
        :return:
        """
        from datetime import datetime
        nextRun = datetime.now().timestamp() + self.cfg.getPingInterval()
        self.sendHeartbeat()
        self.scheduleNextHeartbeat(nextRun)

    def sendHeartbeat(self):
        """
        Posts the current state to the server.
        :param serverURL: the URL to ping.
        :return:
        """
        for name, md in self.cfg.recordingDevices.items():
            try:
                data = marshal(md, recordingDeviceFields)
                data['serviceURL'] = self.cfg.getServiceURL() + API_PREFIX + '/devices/' + name
                targetURL = self.serverURL + API_PREFIX + '/devices/' + name
                logger.info("Pinging " + targetURL)
                resp = self.httpclient.put(targetURL, json=data)
                if resp.status_code != 200:
                    logger.warning("Unable to ping server at " + targetURL + " with " + str(data.keys()) +
                                   ", response is " + str(resp.status_code))
                else:
                    logger.info("Pinged server at " + targetURL + " with " + str(data.items()))
            except:
                logger.exception("Unable to ping server")

    def scheduleNextHeartbeat(self, nextRun):
        """
        Schedules the next ping.
        :param nextRun: when we should run next.
        :param serverURL: the URL to ping.
        :return:
        """
        import threading
        from datetime import datetime
        tilNextTime = max(nextRun - datetime.now().timestamp(), 0)
        logging.getLogger('recorder').info("Scheduling next ping in " + str(round(tilNextTime, 3)) + " seconds")
        threading.Timer(tilNextTime, self.ping).start()



import copy
import faulthandler
import logging.handlers

from flask import Flask
from flask_restful import Api

from core.handler import AsyncHandler
from core.httpclient import RequestsBasedHttpClient
from core.interface import API_PREFIX
from core.reactor import Reactor
from recorder.common.config import Config
from recorder.common.heartbeater import Heartbeater
from recorder.resources.measurements import Measurements, Measurement, AbortMeasurement
from recorder.resources.recordingdevices import RecordingDevices, RecordingDevice, SelfTest

cfg = Config()
reactor = Reactor()
httpclient = RequestsBasedHttpClient()
heartbeater = Heartbeater(httpclient, cfg)
inject = {
    'recordingDevices': cfg.recordingDevices,
    'measurements': cfg.measurements,
    'reactor': reactor,
    'heartbeater': heartbeater
}
app = Flask(__name__)
api = Api(app)

# register a thread dumper
faulthandler.enable()
if hasattr(faulthandler, 'register'):
    import signal

    faulthandler.register(signal.SIGUSR2, all_threads=True)

# GET: the recordingDevices on this host
api.add_resource(RecordingDevices, API_PREFIX + '/devices', resource_class_kwargs=inject)
# GET: the current state of this particular device
# PATCH: mutate specific aspects of configuration
api.add_resource(RecordingDevice, API_PREFIX + '/devices/<deviceId>', resource_class_kwargs=inject)
# GET: triggers a self test and returns the results
api.add_resource(SelfTest, API_PREFIX + '/devices/<deviceId>/selftest', resource_class_kwargs=inject)
# GET: the measurements made by this device in this session
api.add_resource(Measurements, API_PREFIX + '/devices/<deviceId>/measurements', resource_class_kwargs=inject)
# GET: the state of this measurement
# PUT: create a new measurement (duration, start time)
# DELETE: remove the named measurement
api.add_resource(Measurement, API_PREFIX + '/devices/<deviceId>/measurements/<measurementId>', resource_class_kwargs=inject)
# GET: triggers a stop of the named measurement
api.add_resource(AbortMeasurement, API_PREFIX + '/devices/<deviceId>/measurements/<measurementId>/abort',
                 resource_class_kwargs=inject)


def wireHandlers(cfg):
    """
    If the device is configured to run against a remote server, ping that device on a scheduled basis with our current
    state.
    :param cfg: the config object.
    :return:
    """
    logger = logging.getLogger('recorder')
    httpPoster = cfg.handlers.get('remote')
    csvLogger = cfg.handlers.get('local')
    activeHandler = None
    if httpPoster is None:
        if csvLogger is None:
            logger.warning("App is running with discard handler only, ALL DATA WILL BE DISCARDED!!!")
        else:
            logger.info("App is running in standalone mode, logging data to local filesystem")
            activeHandler = csvLogger
    else:
        logger.info("App is running against remote server, logging data to " + httpPoster.target)
        activeHandler = httpPoster
        heartbeater.serverURL = httpPoster.target
        heartbeater.ping()

    if activeHandler is not None:
        for device in cfg.recordingDevices.values():
            if activeHandler is httpPoster:
                httpPoster.deviceName = device.name
            copied = copy.copy(activeHandler)
            device.dataHandler = copied if not cfg.useAsyncHandlers else AsyncHandler('recorder', copied)


def main(args=None):
    """ The main routine. """
    cfg.configureLogger()
    wireHandlers(cfg)
    # get config from a flask standard place not our config yml
    app.run(debug=cfg.runInDebug(), host='0.0.0.0', port=cfg.getPort())


if __name__ == '__main__':
    main()



from setuptools import setup

from versioneer import get_cmdclass, get_version

def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='vibe-recorder',
      version=get_version(),
      cmdclass=get_cmdclass(),
      description='A vibration analysis and data acquisition suite for the rpi',
      long_description=readme(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Framework :: Flask',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Intended Audience :: End Users/Desktop'
      ],
      url='http://github.com/3ll3d00d/vibe',
      author='Matt Khan',
      author_email='mattkhan+vibe@gmail.com',
      license='MIT',
      packages=[
          'recorder',
          'recorder.common',
          'recorder.resources',
          'core'
      ],
      entry_points={
          'console_scripts': [
              'recorder = recorder.app:main',
          ],
      },
      install_requires=[
          'smbus2',
          'flask',
          'flask-restful',
          'pyyaml',
          'requests'
      ],
      setup_requires=[
          'pytest-runner',
          'versioneer'
      ],
      tests_require=[
          'pytest'
      ],
      include_package_data=True,
      zip_safe=False)



import abc
import logging
import sys
import time

from core.handler import Discarder
from core.interface import RecordingDeviceStatus

SAMPLE_TIME = 'time'
ACCEL_X = 'ac_x'
ACCEL_Y = 'ac_y'
ACCEL_Z = 'ac_z'
GYRO_X = 'gy_x'
GYRO_Y = 'gy_y'
GYRO_Z = 'gy_z'
TEMP = 'temp'

logger = logging.getLogger('recorder.accelerometer')


class Accelerometer(object):
    """
    A simple base class to represent an accelerometer, exists to enable the system to be tested in the absence of a
    physical device.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, fs=None, samplesPerBatch=None, dataHandler=None):
        """
        Initialises the accelerometer to use a default sample rate of 500Hz, a samplesPerBatch that accommodates 1/4s
        worth of data and a dataHandler function that simply logs data to the screen.
        :param: fs: the sample rate
        :param: samplesPerBatch: the number of samples that each provideData block should yield.
        :param: dataHandler: a function that accepts the data produced by initialiseDevice and does something with it.
        """
        if fs is None:
            self.fs = 500
        else:
            self.fs = fs
        if samplesPerBatch is None:
            self.samplesPerBatch = self.fs / 4
        else:
            self.samplesPerBatch = samplesPerBatch
        if dataHandler is None:
            self.dataHandler = Discarder()
        else:
            self.dataHandler = dataHandler
        self.breakRead = False
        self.startTime = 0
        self.failureCode = None
        self.measurementOverflowed = False
        self._sampleIdx = 0
        self.status = RecordingDeviceStatus.NEW

    def doInit(self):
        try:
            logger.info("Initialising device")
            self.initialiseDevice()
            self.status = RecordingDeviceStatus.INITIALISED
            logger.info("Initialisation complete")
        except:
            self.status = RecordingDeviceStatus.FAILED
            self.failureCode = str(sys.exc_info())
            logger.exception("Initialisation failure")

    def start(self, measurementId, durationInSeconds=None):
        """
        Initialises the device if required then enters a read loop taking data from the provider and passing it to the
         handler. It will continue until either breakRead is true or the duration (if provided) has passed.
        :return:
        """
        logger.info(">> measurement " + measurementId +
                    ((" for " + str(durationInSeconds)) if durationInSeconds is not None else " until break"))
        self.failureCode = None
        self.measurementOverflowed = False
        self.dataHandler.start(measurementId)
        self.breakRead = False
        self.startTime = time.time()
        self.doInit()
        # this must follow doInit because doInit sets status to INITIALISED
        self.status = RecordingDeviceStatus.RECORDING
        elapsedTime = 0
        try:
            self._sampleIdx = 0
            while True:
                logger.debug(measurementId + " provideData ")
                self.dataHandler.handle(self.provideData())
                elapsedTime = time.time() - self.startTime
                if self.breakRead or durationInSeconds is not None and elapsedTime > durationInSeconds:
                    logger.debug(measurementId + " breaking provideData")
                    self.startTime = 0
                    break
        except:
            self.status = RecordingDeviceStatus.FAILED
            self.failureCode = str(sys.exc_info())
            logger.exception(measurementId + " failed")
        finally:
            expectedSamples = self.fs * (durationInSeconds if durationInSeconds is not None else elapsedTime)
            if self._sampleIdx < expectedSamples:
                self.status = RecordingDeviceStatus.FAILED
                self.failureCode = "Insufficient samples " + str(self._sampleIdx) + " for " + \
                                   str(elapsedTime) + " second long run, expected " + str(expectedSamples)
            self._sampleIdx = 0
            if self.measurementOverflowed:
                self.status = RecordingDeviceStatus.FAILED
                self.failureCode = "Measurement overflow detected"
            if self.status == RecordingDeviceStatus.FAILED:
                logger.error("<< measurement " + measurementId + " - FAILED - " + self.failureCode)
            else:
                self.status = RecordingDeviceStatus.INITIALISED
                logger.info("<< measurement " + measurementId + " - " + self.status.name)
            self.dataHandler.stop(measurementId, self.failureCode)
            if self.status == RecordingDeviceStatus.FAILED:
                logger.warning("Reinitialising device after measurement failure")
                self.doInit()

    def signalStop(self):
        """
        Signals the accelerometer to stop reading after the next read completes.
        :return:
        """
        logger.info("Signalling stop")
        self.breakRead = True

    @abc.abstractmethod
    def provideData(self):
        """
        reads the underlying device to provide a batch of raw data.
        :return: a list of data where each item is a single sample of data converted into real values and stored as a
        dict.
        """
        pass

    @abc.abstractmethod
    def initialiseDevice(self):
        """
        initialises the underlying device
        """
        pass
